/* GENIUS Calculator
 * Copyright (C) 1997-2003 Jiri (George) Lebl
 *
 * Author: Jiri (George) Lebl
 *
 * This file is part of Genius.
 *
 * Genius is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __LEXER_H__
#define __LEXER_H__

/* If you wish to use a file pointer with genius you must first
   _open it, and you should close it with the _close function below */
void gel_lexer_open (FILE *fp);
void gel_lexer_parse (FILE *fp);
void gel_lexer_close (FILE *fp);

#endif
